from .openai import OpenAI
from .replicate import Replicate
from .sql import Sql
from .substation import SubstationTask
from .vectordb import Store

__all__ = ("SubstationTask", "OpenAI", "Store", "Replicate", "Sql")



from typing import Any, Dict, List, Optional

from psycopg2.extras import Json

from .sql_executor import SqlExecutor


class RequestDataSource:
    def __init__(
        self, database: str, host: str, username: str, password: str, port: int = 5432
    ) -> None:
        self.executor = SqlExecutor(database, host, username, password, port)

    def save_request(self, id: str, batch_count: int) -> int:
        return self.executor.insert(
            """ INSERT INTO requests
                        (id, batch_count)
                        VALUES
                        (%s, %s)""",
            [id, batch_count],
        )

    def save_result(self, request_id: str, order: int, output: Any) -> int:
        return self.executor.insert(
            """ INSERT INTO results
                        (request_id, original_order, output)
                        VALUES
                        (%s, %s, %s)""",
            [request_id, order, Json(output)],
        )

    def get_request_status(self, request_id: str) -> Optional[Any]:
        row = self.executor.fetch_one(
            """
          SELECT r.id, r.batch_count,
              (SELECT COUNT(*) FROM results rs WHERE rs.request_id = r.id) AS result_count
          FROM requests r
          WHERE r.id = %s;
        """,
            [request_id],
        )
        result = None
        if row:
            result = {"request_id": row[0], "batch_count": row[1], "result_count": row[2]}
        return result

    def get_request_results(self, request_id: str) -> List[Dict[str, Any]]:
        rows = self.executor.fetch_all(
            """
          SELECT output
          FROM results
          WHERE request_id = %s
          ORDER BY original_order ASC;
        """,
            [request_id],
        )

        results = []
        for row in rows:
            results.append(row[0])

        return results



from typing import Any, Callable, Dict, List, Optional, Tuple

from langchain.chains import ConversationalRetrievalChain
from langchain.document_loaders import PyPDFLoader
from langchain.embeddings.base import Embeddings
from langchain.text_splitter import RecursiveCharacterTextSplitter

from ...integrations.langchain import SeaplaneLLM, langchain_vectorstore, seaplane_embeddings
from ...logging import log
from ...model.vector import Vector
from ...vector import vector_store


class Store:
    def __init__(self, index: str) -> None:
        self.index = index
        self.chat_history_file: Dict[str, List[Tuple[str, str]]] = {}
        self.text_splitter = RecursiveCharacterTextSplitter(
            chunk_size=1000,
            chunk_overlap=100,
            length_function=len,
            add_start_index=True,
        )        

    def save(
        self, file_name: str, file_url: str, embeddings: Embeddings = seaplane_embeddings
    ) -> None:
        loader = PyPDFLoader(file_url)
        pages = loader.load_and_split()
        texts = self.text_splitter.split_documents(pages)

        embed_vectors = embeddings.embed_documents([page.page_content for page in texts])
        vectors = [
            Vector(
                vector=vector,
                metadata={
                    "page_content": texts[idx].page_content,
                    "metadata": texts[idx].metadata,
                },
            )
            for idx, vector in enumerate(embed_vectors)
        ]

        result = vector_store.insert(self.index, vectors)
        log.info(f"⏳ Saving file {file_name}: {result}")

    def query(
        self, file_name: str, query: str, embeddings: Embeddings = seaplane_embeddings
    ) -> Dict[str, Any]:
        vectorstore = langchain_vectorstore(self.index, embeddings)

        qa = ConversationalRetrievalChain.from_llm(
            llm=SeaplaneLLM(),
            retriever=vectorstore.as_retriever(),
            return_source_documents=True,
        )

        history = self.chat_history_file.get(file_name, None)
        if history is None:
            history = []
            self.chat_history_file[file_name] = []

        result = qa({"question": query, "chat_history": history})
        self.chat_history_file[file_name].append((query, result["answer"]))

        return {"answer": result["answer"], "history": history}


class VectorDbTask:
    def __init__(self, func: Callable[[Any], Any], id: str, model: Optional[str]) -> None:
        self.func = func
        self.args: Optional[Tuple[Any, ...]] = None
        self.kwargs: Optional[Dict[str, Any]] = None
        self.type = "vectordb"
        self.model = model
        self.id = id

    def process(self, *args: Any, **kwargs: Any) -> Any:
        self.args = args
        self.kwargs = kwargs

        if self.type == "vectordb":
            log.info("Accessing Vector DB task...")
            # self.args = self.args + (Store(),)

            return self.func(*self.args, **self.kwargs)

    def print(self) -> None:
        log.info(f"id: {self.id}, type: {self.type}, model: {self.model}")



"""process module hooks up to input/output FIFOs.

The default FIFOs are input.fifo (read from) and output.fifo (write
to). These can be overridden with `SEAPLANE_INPUT_FIFO` and
`SEAPLANE_OUTPUT_FIFO` environment variables respectively.

Usage::

    from seaplane.carrier import processor

    # Reverse all messages.
    while True:
        msg = processor.read()  # Read an entire message (bytes).
        processor.write(msg[::-1])  # Write the reversed bytes.
        processor.flush()  # Flush the messages.

"""
from io import BufferedReader, BufferedWriter
import logging
import os
import threading
import time
from typing import List, Optional


class ProcessorIO:
    """ProcessorIO handles IO to a stdpipe with a framed transport.

    ProcessorIO provides three thread-safe methods, `read`, `write` and `flush`.

    `read` reads a single message from `input_fifo`.

    `write` appends a message to be written out, to a `messages` buffer.

    `flush` flushes the messages buffer to the `output_fifo`.
    """

    def __init__(self, input_fifo_path: str, output_fifo_path: str) -> None:
        self._input_fifo_path = input_fifo_path
        self._output_fifo_path = output_fifo_path
        self._input_fifo: Optional[BufferedReader] = None
        self._output_fifo: Optional[BufferedWriter] = None
        self._read_lock = threading.Lock()
        self._write_lock = threading.Lock()
        self._messages: List[bytes] = []

    def read(self) -> bytes:
        """
        `read` takes a single message, the frame format is a 4 byte header
        (unsigned 32 bit int, in big-endian order) indicating the frame size,
        followed by the frame itself.

        Example: ```b'\\x00\\x00\\x00\\x05hello'```
        """
        with self._read_lock:
            if self._input_fifo is not None:
                # Read frame header.
                header = self._input_fifo.read(4)
                frame_size = int.from_bytes(header, "big", signed=False)
                return self._input_fifo.read(frame_size)

        return bytes(0)

    def write(self, msg: bytes) -> None:
        """
        `write` appends a message to the message buffer.

        Once all the messages are written, they need to be `flush`ed.
        """
        self._messages.append(msg)

    def flush(self) -> None:
        """
        `flush` flushes the messages buffer to the output_fifo, using a format:
        - 4 byte header (unsigned 32 bit int) with the frame size
        - 4 byte header (unsigned 32 bit int) with the number of messages
        - all the messages in the frame format:
          - 4 byte header (unsigned 32 bit int) with the message size
          - message bytes

        Example:

        2 messages: `hello` and `world` are encoded as follows:

        ```
        frame_size = b'\\x00\\x00\\x00\\x16' # 22 bytes that follow
        number_of_messages = b'\\x00\\x00\\x00\\x02' # 2 messages
        message1 = b'\\x00\\x00\\x00\\x05hello'
        message2 = b'\\x00\\x00\\x00\\x05world'
        ```

        Which results in this message being flushed to fifo_output:

        ```
        b'\\x00\\x00\\x00\\x16\\x00\\x00\\x00\\x02\\x00\\x00\\x00\\x05hello\\x00\\x00\\x00\\x05world'
        ```

        """
        with self._write_lock:
            output = bytearray()
            num_msgs = len(self._messages).to_bytes(4, "big")
            output.extend(num_msgs)
            for msg in self._messages:
                msg_len = len(msg).to_bytes(4, "big")
                output.extend(msg_len + msg)
            frame_size = len(output).to_bytes(4, "big")
            if self._output_fifo:
                self._output_fifo.write(frame_size + output)
                self._output_fifo.flush()
            self._messages.clear()

    def start(self) -> None:
        logging.info("opening input {}".format(self._input_fifo_path))
        logging.info("opening output {}".format(self._output_fifo_path))
        for _ in range(10):
            try:
                self._input_fifo = open(self._input_fifo_path, "rb")
                self._output_fifo = open(self._output_fifo_path, "wb")
            except FileNotFoundError as exc:
                logging.error(exc)
                time.sleep(2)
            else:
                return
        raise FileNotFoundError("input/output")


processor = ProcessorIO(
    os.getenv("SEAPLANE_INPUT_FIFO", "input.fifo"),
    os.getenv("SEAPLANE_OUTPUT_FIFO", "output.fifo"),
)



from typing import Any, Dict, List, NamedTuple


class CreatedDatabase(NamedTuple):
    """
    Created database.
    """

    name: str
    username: str
    password: str


def to_created_database(database: Dict[str, Any]) -> CreatedDatabase:
    return CreatedDatabase(
        name=database["database"],
        username=database["username"],
        password=database["password"],
    )


def to_list_databases(databases: Dict[str, Any]) -> List[str]:
    return [database["database"] for database in databases["databases"]]



from typing import Any, Callable, Dict, Optional, Tuple

from ...api.api_substation import SubstationAPI
from ...configuration import config
from ...logging import log

substation = SubstationAPI(config)


def substation_injection() -> Callable[[Dict[str, Any]], Any]:
    def model(input: Dict[str, Any]) -> Any:
        return substation.predict(input["prompt"], input["max_output_length"])

    return model


class SubstationTask:
    def __init__(self, func: Callable[[Any], Any], id: str, model: Optional[str]) -> None:
        self.func = func
        self.args: Optional[Tuple[Any, ...]] = None
        self.kwargs: Optional[Dict[str, Any]] = None
        self.type = "inference"
        self.model = model
        self.id = id

    def process(self, *args: Any, **kwargs: Any) -> Any:
        self.args = args
        self.kwargs = kwargs

        if self.model == "mpt-30b":
            log.info("Processing MPT-30B Model...")
            self.args = self.args + (substation_injection(),)

            return self.func(*self.args, **self.kwargs)

    def print(self) -> None:
        log.info(f"id: {self.id}, type: {self.type}, model: {self.model}")



from enum import Enum
from typing import Any, Dict, List, NamedTuple, Optional
import uuid


class Vector(NamedTuple):

    vector: List[float]
    id: str = str(uuid.uuid4())
    metadata: Optional[Dict[Any, Any]] = None


Vectors = List[Vector]


class Distance(Enum):
    COSINE = "cosine"
    DOT = "dot"
    EUCLID = "euclid"

    def __str__(self) -> str:
        return str(self.value)



import json
from typing import Any

import requests

from ..configuration import Configuration, config
from ..util import unwrap
from .api_http import headers
from .api_request import provision_req


class SubstationAPI:
    """
    Class for handle Substation API calls.
    """

    def __init__(self, configuration: Configuration = config) -> None:
        self.url = f"{configuration.substation_endpoint}/completions"
        self.req = provision_req(configuration._token_api)

    # This is the template that works best with the mpt-30b-instruct model;
    # strongly advised to use this format.
    def _format_prompt(self, instruction: str) -> str:
        template = "Below is an instruction that describes a task. Write a response that appropriately completes the request.\n\n###Instruction\n{instruction}\n\n### Response\n"
        return template.format(instruction=instruction)

    def predict(self, prompt: str, max_output_length: int = 3000) -> Any:
        result = unwrap(
            self.req(
                lambda access_token: requests.post(
                    self.url,
                    headers=headers(access_token),
                    json={"max_tokens": max_output_length, "prompt": self._format_prompt(prompt)},
                )
            )
        )

        if (
            type(result) is str
        ):  # Substation FIX, It needs to returns content-type: application/json
            return json.loads(result)
        else:
            return result
